/**
 * A screen object.
 * @class
 * @param {object} args Construction values for public properties.
 */

punx.Sprite = function() {

    /**
     * Sprite sheet to draw from.
     * @public
     * @type Image
     */
    this.sheet = null;

    /**
     * Sprite's X crop offset in the sprite sheet.
     * @public
     * @type int
     * @defaults 0
     */
    this.x = 0;

    /**
     * Sprite's Y crop offset in the sprite sheet.
     * @public
     * @type int
     * @defaults 0
     */
    this.y = 0;

    /**
     * Sprite's crop width in the sprite sheet.
     * @public
     * @type int
     * @defaults 64
     */
    this.w = 64;

    /**
     * Sprite's crop height in the sprite sheet.
     * @public
     * @type int
     * @defaults 64
     */
    this.h = 64;

    /**
     * Sprite draws itself to the context and position supplied.
     * If the sheet isn't set, a red rectangle is drawn.
     * @public
     * @param ctx Canvas context.
     */
    this.draw = function(ctx,x,y,w,h) {
        if (this.sheet !== null) {
            ctx.drawImage(this.sheet,this.x,this.y,this.w,this.h,x,y,w,h);
        }
        else {
            ctx.fillStyle = '#f00';
            ctx.drawRect(x,y,w,h);
        }
    };

    var bitmask = {};
    /**
     * Returns a 2D array, row -> 32-bit chunks of non-transparent pixel indicators, right-padded with zeroes.
     * Bitmasks are cached as per the current x,y of the sprite's crop information.
     */
    this.getBitmask = function() {
        if (bitmask[this.x] == undefined) bitmask[this.x] = {};
        if (bitmask[this.x][this.y] != undefined) return bitmask[this.x][this.y];
        bitmask[this.x][this.y] = [];
        Punx.ictx.clearRect(0,0,this.w,this.h);
        this.draw(Punx.ictx,0,0,this.w,this.h);
        var imageData = Punx.ictx.getImageData(0,0,this.w,this.h);
        for (var i=0; i<this.h; i++) {
            bitmask[this.x][this.y][i] = [];
            var rowStart = i*this.w;
            var chunk = 0;
            var chunkOffset = 0;
            for (var j=0; j<this.w; j++) {
                var bit = (imageData.data[rowStart+j+3] == 0 ? 0 : 1);
                chunkOffset = j % 32;
                chunk |= bit << chunkOffset;
                if (chunkOffset == 31) {
                    bitmask[this.x][this.y][i].append(chunk);
                    chunk = 0;
                }
            }
            if (chunkOffset != 31) {
                chunk <<= (31 - chunkOffset);
                bitmask[this.x][this.y][i].append(chunk);
            }
        }
        return bitmask[this.x][this.y];
    };
};
