/**
 * @author magicmarkers
 * @description A derpy HTML5/JS engine for side-scrolling brawler / beat 'em up / punchy-fighty games.
 * @see https://github.com/magicmarkers/Punx
 */

window.requestAnimationFrame = window.requestAnimationFrame || window.msRequestAnimationFrame || window.mozRequestAnimationFrame || window.webkitRequestAnimationFrame || window.oRequestAnimationFrame || function(cb){
    setTimeout(cb,1000/60);
};

var punx = {}:

/**
 * Determines if two boxes intersect.
 * @public
 * @param {Punx.Box} a
 * @param {Punx.Box} b
 * @returns boolean
 */
punx.aabb = function(a,b) {
    return (!(
        a.x + a.w < b.x || // a too left
        a.x > b.x + b.w || // a too right
        a.y + a.h < b.y || // a too bg (y)
        a.y > b.y + b.h    // a too fg (y)
    ));
};

/**
 * Returns a box describing the intersection of two boxes.
 * This may return strange things if they aren't actually intersecting. Use AABB() first.
 * @public
 * @param {Punx.Box} a
 * @param {Punx.Box} b
 * @returns Punx.Box
 */
punx.aabbc = function(a,b) {
    var c = {
        x : (a.x < b.x ? b.x : a.x),
        y : (a.y < b.y ? b.y : a.y)
    };
    c.w = (a.x + a.w < b.x + b.w ? a.x + a.w : b.x + b.w) - c.x;
    c.h = (a.y + a.h < b.y + b.h ? a.y + a.h : b.y + b.h) - c.y;
    return new Punx.Box(c);
};

/**
 * Returns an array of sprites, sorted according to their proximity to the origin x.
 * @param {number} x Integer x origin point
 * @param {Punx.Sprite[]} sprites Sprites to sort
 * @returns {Punx.Sprite[]} Sorted array
 */
punx.xSort = function(x,sprites) {
    return sprites.sort(function(a,b) {
        var A = (a.x + a.w < x ? x - (a.x + a.w) : a.x - x);
        var B = (b.x + b.w < x ? x - (b.x + b.w) : b.x - x);
        return (A < B ? -1 : (A > B ? 1 : 0));
    });
};

/**
 * Returns an array of sprites, sorted according to their proximity to the origin y.
 * @param {number} y Integer y origin point
 * @param {Punx.Sprite[]} sprites Sprites to sort
 * @returns {Punx.Sprite[]} Sorted array
 */
punx.ySort = function(y,sprites) {
    return sprites.sort(function(a,b) {
        var A = (a.y + a.h < y ? y - (a.y + a.h) : a.y - y);
        var B = (b.y + b.h < y ? y - (b.y + b.h) : b.y - y);
        return (A < B ? -1 : (A > B ? 1 : 0));
    });
};

punx.zSort = function(a,b) {
    return a.z < b.z ? -1 : (b.z > a.z ? 1 : 0);
};


var Punx = new function() {

    /**
     * Integer canvas width.
     * @public
     * @type number
     * @default 960
     */
    this.width = 960;

    /**
     * Integer canvas height.
     * @public
     * @type number
     * @default 540
     */
    this.height = 540;

    this.screen = null;

    var loop = null;

    this.ctx = null;
    this.ictx = null;

    this.draw = function() {
        this.screen.draw();
        window.requestAnimationFrame(this.draw);
    }.bind(this);

    var then = null;
    var update = function() {
        var now = Date.now();
        var delta = (then == null ? 0 : (now - then) / 1000);
        then = now;
        try {
            this.screen.update(delta);
        }
        catch (x) {
            console.log(x.stack);
            window.clearInterval(loop);
        }
    }.bind(this);

    this.init = function(containerElement) {
        var canvas = document.createElement('canvas');
        canvas.setAttribute('width',this.width);
        canvas.setAttribute('height',this.height);
        canvas.setAttribute('tabindex',0);
        canvas.addEventListener('keydown',Punx.Key.onkeydown,true);
        canvas.addEventListener('keyup',Punx.Key.onkeyup,true);
        this.ctx = canvas.getContext('2d');
        this.ctx.imageSmoothingEnabled = false;
        this.ctx.mozImageSmoothingEnabled = false;
        this.ctx.webkitImageSmoothingEnabled = false;
        (containerElement||document.body).appendChild(canvas);
        canvas.focus();

        // imaginary canvas
        var icanvas = document.createElement('canvas');
        icanvas.setAttribute('width',this.width);
        icanvas.setAttribute('height',this.height);
        this.ictx = icanvas.getContext('2d');
        this.ictx.imageSmoothingEnabled = false;
        this.ictx.mozImageSmoothingEnabled = false;
        this.ictx.webkitImageSmoothingEnabled = false;

        loop = window.setInterval(update,1000/60);
        this.draw();
    }

    this.clear = function() {
        this.ctx.clearRect(0,0,this.width,this.height);
    };

};
