/**
 * A screen object.
 * @class
 * @param {object} args Construction values for public properties.
 */

punx.Sprite = function() {

    /**
     * Sprite sheet to draw from.
     * @public
     * @type Image
     */
    this.sheet = null;

    /**
     * Sprite's X crop offset in the sprite sheet.
     * @public
     * @type int
     * @defaults 0
     */
    this.x = 0;

    /**
     * Sprite's Y crop offset in the sprite sheet.
     * @public
     * @type int
     * @defaults 0
     */
    this.y = 0;

    /**
     * Sprite's crop width in the sprite sheet.
     * @public
     * @type int
     * @defaults 64
     */
    this.w = 64;

    /**
     * Sprite's crop height in the sprite sheet.
     * @public
     * @type int
     * @defaults 64
     */
    this.h = 64;

    /**
     * Sprite draws itself to the context and position supplied.
     * If the sheet isn't set, a red rectangle is drawn.
     * @public
     * @param ctx Canvas context.
     */
    this.draw = function(ctx,x,y,w,h) {
        if (this.sheet !== null) {
            ctx.drawImage(this.sheet,this.x,this.y,this.w,this.h,x,y,w,h);
        }
        else {
            ctx.fillStyle = '#f00';
            ctx.drawRect(x,y,w,h);
        }
    };

};

Punx.Sprite.sort = function(a,b) {
    // sorting function that prioritizes sprites with more depth
    // this is mainly used when figuring out the drawing order.
    if (a instanceof Punx.Sprite.Wall ^ b instanceof Punx.Sprite.Wall) return (a instanceof Punx.Sprite.Wall ? -1 : 1); // walls always behind
    else if (a.z != b.z) return (a.z < b.z ? -1 : 1);
    else return (a.y + a.h < b.y + b.h ? -1 : 1); // sprite's bottom is further away
};

// under construction

Audio.prototype.fadeTo = function(fadeTo,callback) {
    if (fadeTo < 0) fadeTo = 0;
    else if (fadeTo > 1) fadeTo = 1;
    var that = this;
    var setVolume = function(setTo) {
        return function() {
            that.volume = setTo;
            if (setTo == fadeTo && callback) callback();
        };
    };
    var delay = 0;
    if (this.volume < fadeTo) {
        for (var v=this.volume; v<=fadeTo; v+=0.1) {
            setTimeout(setVolume(v),delay);
            delay+=100;
        }
    }
    else if (volume > fadeTo) {
        for (var v=this.volume; v>=fadeTo; v-=0.1) {
            setTimeout(setVolume(v),delay);
            delay+=100;
        }
    }
    else if (callback) callback();
};

punx.Sound = new function() {

    this.music = new Audio(); // swap out for one in Library

    this.muted = false;
    this.mute = function() {
        this.muted ^= 1;
        this.music.muted = this.muted;
    };

};

/**
 * Represents a collection of sprites to be drawn to the canvas. Only the active screen is drawn.
 * @class
 * @param {object} args Construction values for public properties.
 */

punx.Screen = function() {

    /**
     * List of {punx.Entity} instances. Should be pre-sorted by the draw order (z-index).
     * @type Punx.Sprites[]
     * @default []
     */
    this.bgs = [];

    this.actors = [];

    this.overlays = [];

    var updated = false;

    /**
     * Update all sprites and remember if updated.
     * Called by the {Punx} update loop.
     */
    this.update = function() {
        for (var i in this.bgs) {
            updated |= this.bgs[i].update();
        }
        for (var i in this.actors) {
            updated |= this.actors[i].update();
        }
        for (var i in this.overlays) {
            updated |= this.overlays[i].update();
        }
    };

    /**
     * Stage back.
     */
    this.back = 0;

    /**
     * Stage front.
     */
    this.front = 0;

    /**
     * Camera tracking X offset.
     */
    this.track = 0;

    /**
     * Draws only if updated.
     * Draws all entities in order.
     * @public
     */
    this.draw = function() {
        if (!updated) return;
        updated = false;
        Punx.clear();
        for (var i in this.bgs) {
            this.bgs[i].draw();
        }
        for (var i in this.actors) {
            this.actors.draw();
        }
        for (var i in this.overlays) {
            this.overlays[i].draw();
        }
    };

    this.move = function(a,x,y) {
        // TODO account for clipping and collision
        a.x = x;
        a.y = y;
    }

};

/**
 * Keyboard interface.
 */
punx.key = new function() {

    /**
     * Object keyed by button name and its current press-state, so that sprites et al. can poll the keyboard.
     * @private
     * @type object<string,boolean>
     */
    var state = {};

    /**
     * Internal map between keyCode and button name.
     * @private
     * @type object<integer,string>
     */
    var map = {
        9:'tab',
        13:'enter',
        27:'escape',
        32:'space',
        37:'left',
        38:'up',
        39:'right',
        40:'down'
    };
    for (var i=48;i<=57;i++) { map[i] = String.fromCharCode(i); } // numbers
    for (var i=97;i<=122;i++) { map[i] = String.fromCharCode(i); } // letters

    this.pressed = function(name) {
        return name in state && state[name];
    };

    /**
     * Document's onkeydown listener, attached to the document by the grid during create().
     * Sets the key's state to true. Keypresses that are combined with special keys like CTRL, ALT, Shift, Meta/Windows, Menu, etc are ignored.
     * @public
     * @param {event} [e=window.event]
     */
    this.onkeydown = function(e) {
        e=e||window.event;
        if (!(e.keyCode in map) || e.altGraphKey || e.metaKey || e.altKey || e.shiftKey || e.ctrlKey) return;
        e.preventDefault();
        e.stopPropagation();
        state[map[e.keyCode]] = true;
    }.bind(this);

    /**
     * Document's keyup listener, attached to the document by the grid during create().
     * Sets the key's state to false.
     * @public
     * @param {event} [e=window.event]
     */
    this.onkeyup = function(e) {
        e=e||window.event;
        if (!(e.keyCode in map)) return;
        e.preventDefault();
        e.stopPropagation();
        state[map[e.keyCode]] = false;
    }.bind(this);

};

/**
 * @author magicmarkers
 * @description A derpy HTML5/JS engine for side-scrolling brawler / beat 'em up / punchy-fighty games.
 * @see https://github.com/magicmarkers/Punx
 */

window.requestAnimationFrame = window.requestAnimationFrame || window.msRequestAnimationFrame || window.mozRequestAnimationFrame || window.webkitRequestAnimationFrame || window.oRequestAnimationFrame || function(cb){
    setTimeout(cb,1000/60);
};

var punx = {}:

/**
 * Determines if two boxes intersect.
 * @public
 * @param {Punx.Box} a
 * @param {Punx.Box} b
 * @returns boolean
 */
punx.aabb = function(a,b) {
    return (!(
        a.x + a.w < b.x || // a too left
        a.x > b.x + b.w || // a too right
        a.y + a.h < b.y || // a too bg (y)
        a.y > b.y + b.h    // a too fg (y)
    ));
};

/**
 * Returns a box describing the intersection of two boxes.
 * This may return strange things if they aren't actually intersecting. Use AABB() first.
 * @public
 * @param {Punx.Box} a
 * @param {Punx.Box} b
 * @returns Punx.Box
 */
punx.aabbc = function(a,b) {
    var c = {
        x : (a.x < b.x ? b.x : a.x),
        y : (a.y < b.y ? b.y : a.y)
    };
    c.w = (a.x + a.w < b.x + b.w ? a.x + a.w : b.x + b.w) - c.x;
    c.h = (a.y + a.h < b.y + b.h ? a.y + a.h : b.y + b.h) - c.y;
    return new Punx.Box(c);
};

/**
 * Returns an array of sprites, sorted according to their proximity to the origin x.
 * @param {number} x Integer x origin point
 * @param {Punx.Sprite[]} sprites Sprites to sort
 * @returns {Punx.Sprite[]} Sorted array
 */
punx.xSort = function(x,sprites) {
    return sprites.sort(function(a,b) {
        var A = (a.x + a.w < x ? x - (a.x + a.w) : a.x - x);
        var B = (b.x + b.w < x ? x - (b.x + b.w) : b.x - x);
        return (A < B ? -1 : (A > B ? 1 : 0));
    });
};

/**
 * Returns an array of sprites, sorted according to their proximity to the origin y.
 * @param {number} y Integer y origin point
 * @param {Punx.Sprite[]} sprites Sprites to sort
 * @returns {Punx.Sprite[]} Sorted array
 */
punx.ySort = function(y,sprites) {
    return sprites.sort(function(a,b) {
        var A = (a.y + a.h < y ? y - (a.y + a.h) : a.y - y);
        var B = (b.y + b.h < y ? y - (b.y + b.h) : b.y - y);
        return (A < B ? -1 : (A > B ? 1 : 0));
    });
};

punx.zSort = function(a,b) {
    return a.z < b.z ? -1 : (b.z > a.z ? 1 : 0);
};

punx.Game = function() {

    /**
     * Integer canvas width.
     * @public
     * @type number
     * @default 960
     */
    this.width = 960;

    /**
     * Integer canvas height.
     * @public
     * @type number
     * @default 540
     */
    this.height = 540;

    this.screen = null;

    var loop = null;

    this.ctx = null;
    this.ictx = null;

    this.frame = 0;

    this.draw = function() {
        this.screen.draw();
        window.requestAnimationFrame(this.draw);
    }.bind(this);

    var update = function() {
        try {
            this.screen.update();
            this.frame = ++this.frame % 60;
        }
        catch (x) {
            console.log(x.stack);
            window.clearInterval(loop);
        }
    }.bind(this);

    this.init = function(containerElement) {
        var canvas = document.createElement('canvas');
        canvas.setAttribute('width',this.width);
        canvas.setAttribute('height',this.height);
        canvas.setAttribute('tabindex',0);
        canvas.addEventListener('keydown',Punx.Key.onkeydown,true);
        canvas.addEventListener('keyup',Punx.Key.onkeyup,true);
        this.ctx = canvas.getContext('2d');
        this.ctx.imageSmoothingEnabled = false;
        this.ctx.mozImageSmoothingEnabled = false;
        this.ctx.webkitImageSmoothingEnabled = false;
        (containerElement||document.body).appendChild(canvas);
        canvas.focus();

        // imaginary canvas
        var icanvas = document.createElement('canvas');
        icanvas.setAttribute('width',this.width);
        icanvas.setAttribute('height',this.height);
        this.ictx = icanvas.getContext('2d');
        this.ictx.imageSmoothingEnabled = false;
        this.ictx.mozImageSmoothingEnabled = false;
        this.ictx.webkitImageSmoothingEnabled = false;

        loop = window.setInterval(update,1000/60);
        this.draw();
    }

    this.clear = function() {
        this.ctx.clearRect(0,0,this.width,this.height);
    };

};
punx.entity = {};

/**
 * Entity can be collided by another entity.
 * @interface
 */
punx.entity.Collidable = function() {}:
punx.entity.Collidable.prototype.collide = function(entity) {};

/**
 * Entity is drawable.
 * @interface
 */
punx.entity.Drawable = function() {};
punx.entity.Drawable.prototype.draw = function(ctx,track) {};

/**
 * Screen inhabitants. Includes backgrounds, NPC, baddies, player, text, anything drawn or collidable.
 * @class
 */
punx.Entity = function(x,y) {

    this.x = x;
    this.y = y;

    this.update = function() {
        var updated = false;
        for (var i in this.elements) {
            updated |= this.elements[i].update();
        }
        return updated;
    };

};
// actors are any non-inert stage element. these include baddies, powerups, and props.
// actors interact, such as via collision.

Punx.Sprite.Actor = function(args) {
    Punx.Sprite.call(this,args);

    this.name = args['name']||'Actor';

    this.collidable = true;

    this.busy = false; // set this to disable user input or to denote that a multi-cycle process is taking place and you dont want to be interrupted

    // get collided by another actor, react to it.
    // this is meant to be overloaded.
    // dont modify the other actor, as notification order is undefined. it's only provided for context so you can react appropriately.
    // chances are if its the initial collision then there will be a large swath of calls to this function within a short period of time, as
    // collisions are checked every screen update.
    // set flags etc accordingly to "debounce" collision notifications.
    // for example, when the player is collided by a baddy, inflict self-damage and set a flag and timeout for recover-time invulnerability.
    this.collide = function(actor) {
    };

    this.move = function(stage,direction,amount) {
        amount = amount || 5;
        this.setGroup(direction);
        var x = this.x;
        var y = this.y;
        switch (direction) {
            case 'left' : { x -= amount; break; }
            case 'right' : { x += amount; break; }
            case 'up' : { y -= amount; break; }
            case 'down' : { y += amount; break; }
        }
        if (!stage.clipping(this,x,y)) {
            this.x = x;
            this.y = y;
        }
        return true;
    }

}; Punx.Sprite.Actor.prototype = new Punx.Sprite();

